"""
coincident: Search and analysis of STV Precursor Coincident Datasets
"""

from __future__ import annotations

from coincident.io import download, sliderule, xarray

__all__ = ["sliderule", "xarray", "download"]
